# Generated File
from __future__ import annotations

import dataclasses
import struct
import typing
import typing_extensions

from retro_data_structures import json_util
from retro_data_structures.game_check import Game
from retro_data_structures.properties.base_property import BaseObjectType
from retro_data_structures.properties.field_reflection import FieldReflection
from retro_data_structures.properties.echoes.archetypes.EditorProperties import EditorProperties

if typing.TYPE_CHECKING:
    from retro_data_structures.asset_manager import AssetManager
    from retro_data_structures.base_resource import Dependency

    class PlayerStateChangeJson(typing_extensions.TypedDict):
        editor_properties: json_util.JsonObject
        item_to_change: int
        capacity_increase: int
        amount: int
        command: int
        command_action: int
    

@dataclasses.dataclass()
class PlayerStateChange(BaseObjectType):
    editor_properties: EditorProperties = dataclasses.field(default_factory=EditorProperties, metadata={
        'reflection': FieldReflection[EditorProperties](
            EditorProperties, id=0x255a4580, original_name='EditorProperties', from_json=EditorProperties.from_json, to_json=EditorProperties.to_json
        ),
    })
    item_to_change: int = dataclasses.field(default=0, metadata={
        'reflection': FieldReflection[int](
            int, id=0x4a3491bd, original_name='ItemToChange'
        ),
    })
    capacity_increase: int = dataclasses.field(default=1, metadata={
        'reflection': FieldReflection[int](
            int, id=0x28c71b54, original_name='CapacityIncrease'
        ),
    })
    amount: int = dataclasses.field(default=1, metadata={
        'reflection': FieldReflection[int](
            int, id=0x94af1445, original_name='Amount'
        ),
    })
    command: int = dataclasses.field(default=0, metadata={
        'reflection': FieldReflection[int](
            int, id=0x94ba5737, original_name='Command'
        ),
    })
    command_action: int = dataclasses.field(default=0, metadata={
        'reflection': FieldReflection[int](
            int, id=0x363d98a4, original_name='CommandAction'
        ),
    })

    @classmethod
    def game(cls) -> Game:
        return Game.ECHOES

    def get_name(self) -> str | None:
        return self.editor_properties.name

    def set_name(self, name: str) -> None:
        self.editor_properties.name = name

    @classmethod
    def object_type(cls) -> str:
        return 'PSCH'

    @classmethod
    def from_stream(cls, data: typing.BinaryIO, size: int | None = None, default_override: dict | None = None) -> typing_extensions.Self:
        struct_id, size, property_count = struct.unpack(">LHH", data.read(8))
        assert struct_id == 0xFFFFFFFF
        root_size_start = data.tell() - 2

        if (result := cls._fast_decode(data, property_count)) is not None:
            return result

        present_fields = default_override or {}
        for _ in range(property_count):
            property_id, property_size = struct.unpack(">LH", data.read(6))
            start = data.tell()
            try:
                property_name, decoder = _property_decoder[property_id]
                present_fields[property_name] = decoder(data, property_size)
            except KeyError:
                raise RuntimeError(f"Unknown property: 0x{property_id:08x}")
            assert data.tell() - start == property_size

        assert data.tell() - root_size_start == size
        return cls(**present_fields)

    @classmethod
    def _fast_decode(cls, data: typing.BinaryIO, property_count: int) -> typing_extensions.Self | None:
        if property_count != 6:
            return None
    
        property_id, property_size = struct.unpack(">LH", data.read(6))
        assert property_id == 0x255a4580
        editor_properties = EditorProperties.from_stream(data, property_size)
    
        property_id, property_size = struct.unpack(">LH", data.read(6))
        assert property_id == 0x4a3491bd
        item_to_change = struct.unpack('>l', data.read(4))[0]
    
        property_id, property_size = struct.unpack(">LH", data.read(6))
        assert property_id == 0x28c71b54
        capacity_increase = struct.unpack('>l', data.read(4))[0]
    
        property_id, property_size = struct.unpack(">LH", data.read(6))
        assert property_id == 0x94af1445
        amount = struct.unpack('>l', data.read(4))[0]
    
        property_id, property_size = struct.unpack(">LH", data.read(6))
        assert property_id == 0x94ba5737
        command = struct.unpack('>l', data.read(4))[0]
    
        property_id, property_size = struct.unpack(">LH", data.read(6))
        assert property_id == 0x363d98a4
        command_action = struct.unpack('>l', data.read(4))[0]
    
        return cls(editor_properties, item_to_change, capacity_increase, amount, command, command_action)

    def to_stream(self, data: typing.BinaryIO, default_override: dict | None = None) -> None:
        default_override = default_override or {}
        data.write(b'\xff\xff\xff\xff')  # struct object id
        root_size_offset = data.tell()
        data.write(b'\x00\x00')  # placeholder for root struct size
        data.write(b'\x00\x06')  # 6 properties

        data.write(b'%ZE\x80')  # 0x255a4580
        before = data.tell()
        data.write(b'\x00\x00')  # size placeholder
        self.editor_properties.to_stream(data)
        after = data.tell()
        data.seek(before)
        data.write(struct.pack(">H", after - before - 2))
        data.seek(after)

        data.write(b'J4\x91\xbd')  # 0x4a3491bd
        data.write(b'\x00\x04')  # size
        data.write(struct.pack('>l', self.item_to_change))

        data.write(b'(\xc7\x1bT')  # 0x28c71b54
        data.write(b'\x00\x04')  # size
        data.write(struct.pack('>l', self.capacity_increase))

        data.write(b'\x94\xaf\x14E')  # 0x94af1445
        data.write(b'\x00\x04')  # size
        data.write(struct.pack('>l', self.amount))

        data.write(b'\x94\xbaW7')  # 0x94ba5737
        data.write(b'\x00\x04')  # size
        data.write(struct.pack('>l', self.command))

        data.write(b'6=\x98\xa4')  # 0x363d98a4
        data.write(b'\x00\x04')  # size
        data.write(struct.pack('>l', self.command_action))

        struct_end_offset = data.tell()
        data.seek(root_size_offset)
        data.write(struct.pack(">H", struct_end_offset - root_size_offset - 2))
        data.seek(struct_end_offset)

    @classmethod
    def from_json(cls, data: json_util.JsonValue) -> typing_extensions.Self:
        json_data = typing.cast("PlayerStateChangeJson", data)
        return cls(
            editor_properties=EditorProperties.from_json(json_data['editor_properties']),
            item_to_change=json_data['item_to_change'],
            capacity_increase=json_data['capacity_increase'],
            amount=json_data['amount'],
            command=json_data['command'],
            command_action=json_data['command_action'],
        )

    def to_json(self) -> json_util.JsonObject:
        return {
            'editor_properties': self.editor_properties.to_json(),
            'item_to_change': self.item_to_change,
            'capacity_increase': self.capacity_increase,
            'amount': self.amount,
            'command': self.command,
            'command_action': self.command_action,
        }

    def dependencies_for(self, asset_manager: AssetManager) -> typing.Iterator[Dependency]:
        for method, field_name, field_type in [
            (self.editor_properties.dependencies_for, "editor_properties", "EditorProperties"),
        ]:
            try:
                yield from method(asset_manager)
            except Exception as e:
                raise Exception(
                    f"Error finding dependencies for PlayerStateChange.{field_name} ({field_type}): {e}"
                )


def _decode_item_to_change(data: typing.BinaryIO, property_size: int):
    return struct.unpack('>l', data.read(4))[0]


def _decode_capacity_increase(data: typing.BinaryIO, property_size: int):
    return struct.unpack('>l', data.read(4))[0]


def _decode_amount(data: typing.BinaryIO, property_size: int):
    return struct.unpack('>l', data.read(4))[0]


def _decode_command(data: typing.BinaryIO, property_size: int):
    return struct.unpack('>l', data.read(4))[0]


def _decode_command_action(data: typing.BinaryIO, property_size: int):
    return struct.unpack('>l', data.read(4))[0]


_property_decoder: typing.Dict[int, typing.Tuple[str, typing.Callable[[typing.BinaryIO, int], typing.Any]]] = {
    0x255a4580: ('editor_properties', EditorProperties.from_stream),
    0x4a3491bd: ('item_to_change', _decode_item_to_change),
    0x28c71b54: ('capacity_increase', _decode_capacity_increase),
    0x94af1445: ('amount', _decode_amount),
    0x94ba5737: ('command', _decode_command),
    0x363d98a4: ('command_action', _decode_command_action),
}
